package CFG;

import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import parser.Node;

public class CFGUtil {
	
	static Set<Node> lastStatementsOf(Node root) throws Exception //the last statements that can possibly be executed in root
	{
		if(root.type.equals("prog"))
			return lastStatementsOf(root.children.lastElement());
		else if(root.type.equals("stmt") && (root.rule.equals("while") || root.rule.equals("skip") || root.rule.equals("assignment")))
		{
			Set<Node> ans = new HashSet<>();
			ans.add(root);
			return ans;
		}
		else if(root.type.equals("stmt")) //all types of if
		{
			Set<Node> ans = new HashSet<>();
			for(Node ch:root.children)
				if(ch.type.equals("prog"))
				{
					Set<Node> x = lastStatementsOf(ch);
					ans.addAll(x);
				}
			return ans;
		}
		throw new Exception("Cannot compute lastStatementsOf a non-stmt non-prog Node. The node is:"+root);
	}
	
	@SuppressWarnings("unused")
	public static void addEdges(Node root) throws Exception //recursively adds all the edges in the parse subtree of root
	{
		
		if(root.type.equals("prog")) //add normal-flow edges
		{
			for(int i=1;i<root.children.size();++i)
			{
				Node prevChild = root.children.elementAt(i-1);
				Node curChild = root.children.elementAt(i);
				Set<Node> prevLast = lastStatementsOf(prevChild);
				for(Node x:prevLast)
				{
					CFG.Edge e = new CFG.Edge(x, curChild, "normal-flow");
				}
			}
			
			if(root.parent==null) //add end-program edges
			{
				Set<Node> lastOfThis = lastStatementsOf(root);
				for(Node x:lastOfThis)
				{
					CFG.Edge e = new CFG.Edge(x, null, "normal-flow");
				}
			}
		}
		else if(root.type.equals("stmt") && root.rule.equals("while")) //add while edges (while-to-block, block-to-while)
		{
			Node whileNode = root;
			Node firstInBlock = root.children.elementAt(1).children.firstElement();
			CFG.Edge e = new CFG.Edge(whileNode, firstInBlock, "while-to-block");
			Set<Node> lastInBlock = lastStatementsOf(root.children.elementAt(1));
			for(Node x:lastInBlock)
			{
				CFG.Edge e1 = new CFG.Edge(x, whileNode, "block-to-while");
			}
		}
		else if(root.type.equals("stmt") && root.rule.equals("if")) //add if edges (if-to-then, if-to-else)
		{
			Node ifNode = root;
			Node thenProg = ifNode.children.elementAt(1);
			Node elseProg = ifNode.children.elementAt(2);
			Node firstInThen = thenProg.children.firstElement();
			Node firstInElse = elseProg.children.firstElement();
			CFG.Edge e = new CFG.Edge(ifNode, firstInThen, "if-to-then");
			CFG.Edge e1 = new CFG.Edge(ifNode, firstInElse, "if-to-else");
		}
		else if(root.type.equals("stmt") && root.rule.equals("if*")) //add if* edges (if*-to-then, if*-to-else)
		{
			Node ifNode = root;
			Node thenProg = ifNode.children.elementAt(0);
			Node elseProg = ifNode.children.elementAt(1);
			Node firstInThen = thenProg.children.firstElement();
			Node firstInElse = elseProg.children.firstElement();
			CFG.Edge e = new CFG.Edge(ifNode, firstInThen, "if*-to-then");
			CFG.Edge e1 = new CFG.Edge(ifNode, firstInElse, "if*-to-else");
		}
		else if(root.type.equals("stmt") && root.rule.equals("ifprob")) //add ifprob edges (ifprob-to-then, ifprob-to-else)
		{
			Node ifNode = root;
			Node thenProg = ifNode.children.elementAt(1);
			Node elseProg = ifNode.children.elementAt(2);
			Node firstInThen = thenProg.children.firstElement();
			Node firstInElse = elseProg.children.firstElement();
			CFG.Edge e = new CFG.Edge(ifNode, firstInThen, "ifprob-to-then");
			CFG.Edge e1 = new CFG.Edge(ifNode, firstInElse, "ifprob-to-else");
		}
		
		//recurse
		for(Node ch:root.children)
			addEdges(ch);
	}

	public static TreeSet<Node> getSuccessorsOf(Node stmt) {
		TreeSet<Node> ans = new TreeSet<>();
		for(Edge e:Edge.allEdges)
			if(e.u==stmt)
				ans.add(e.v);
		return ans;
	}

	public static Node getTrueNextOf(Node stmt) throws Exception {
		for(Edge e:Edge.allEdges)
			if(e.u==stmt && (e.type.contains("to-then") || e.type.contains("to-block")))
				return e.v;
		throw new Exception("The following statement has no true next:"+stmt);
	}
	
	public static Node getFalseNextOf(Node stmt) throws Exception
	{
		for(Edge e:Edge.allEdges)
			if(e.u==stmt && !e.type.contains("to-then") && !e.type.contains("to-block"))
				return e.v;
		throw new Exception("The following statement has no false next:"+stmt);
	}
	
}
