package DSM;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import CFG.CFGUtil;
import farkas.farkasUtil;
import net.sf.javailp.Result;
import parser.Node;
import parser.Parser;

public class DSMUtil {
	
	static Map<Node, Poly> eta = new TreeMap<>(Comparator.nullsLast(Comparator.naturalOrder()));
	
	public static int eCnt = 1;
	public static void createEtaTemplates()
	{
		Vector<Node> statements = new Vector<Node>(); 
		statements.addAll(Node.allStatements);
		statements.add(null);
		Set<String> pvars = Node.getAllPvars();
		for(Node stmt:statements)
		{
			//create an eta template for stmt
			Poly etaStmt = new Poly();
			//add one term for each pvar
			for(String pvar:pvars)
			{
				PolyTerm t = new PolyTerm();
				t.constant = 1;
				t.powers.put(pvar, 1);
				String etaVar = "e"+eCnt;
				eCnt++;
				t.powers.put(etaVar, 1);
				etaStmt.addTerm(t);
			}
			//add the constant term
			PolyTerm cns = new PolyTerm();
			cns.constant = 1;
			String cnsEtaVar = "e"+eCnt;
			eCnt++;
			cns.powers.put(cnsEtaVar, 1);
			etaStmt.addTerm(cns);
			
			//add the eta
			eta.put(stmt, etaStmt);
		}
		
		//print the eta
		printEta();
	}
	
	private static void printEta() {
		
		System.err.println("Eta templates:");
		System.err.println("========================================");
		
		for(Node stmt:eta.keySet())
			if(stmt!=null)
				System.err.println("eta("+stmt.statementId+") = "+eta.get(stmt));
			else
				System.err.println("eta(end) = "+eta.get(stmt));
		System.err.println("========================================");
	}

	public static void addDSMConditions() throws Exception
	{
		//add the conditions
		Vector<Node> statements = Node.allStatements;
		for(Node stmt:statements)
		{
			//D1
			if(stmt.rule.equals("assignment") || stmt.rule.equals("skip"))
				addD1(stmt);
			//D2
			if(stmt.rule.equals("while") || stmt.rule.equals("if"))
				addD2(stmt);
			//D3
			if(stmt.rule.equals("if*"))
				addD3(stmt);
			//D4
			if(stmt.rule.equals("ifprob"))
				addD4(stmt);
		}
		//D5
		Node lin = statements.firstElement();
		addD5(lin);
	}
	
	private static void addD1(Node stmt) throws Exception {
		
		System.err.println("Processing D1 conditions @"+stmt.statementId+":");
		System.err.println("========================================");
		
		Node rhs = null, lhs=null;
		
		if(stmt.rule.equals("assignment"))
		{
			rhs = stmt.children.elementAt(1);
			lhs = stmt.children.elementAt(0);
		}
		Node invariant = stmt.invariant;
		Set<Poly> invariantConditions = invariant.getPolySetOfBexpr();
		Set<Poly> allPostConditions = new HashSet<>();
		String lhsVar = lhs!=null?lhs.name:Node.getAllPvars().first();
		Poly rhsPoly = rhs!=null?rhs.getPolyOfExpr():new Poly(Node.getAllPvars().first());
		Node nextStmt = CFG.CFGUtil.getSuccessorsOf(stmt).first();
		
		//add all conditions of type 1
		TreeSet<String> rvars = stmt.getRvars();
		Set<Vector<Integer>> allPossibleValues = Parser.getAllPossibleCombinationsOfValuesOfRvars(rvars);
		
		for(Vector<Integer> rvarValuation:allPossibleValues)
		{
			Poly etaL = eta.get(stmt).deepCopy();
			Poly etaLprime = eta.get(nextStmt).deepCopy();
			etaLprime.replaceVarWithPoly(lhsVar, rhsPoly);
			int i=0;
			for(String rvar:rvars)
			{
				int value = rvarValuation.elementAt(i);
				etaLprime.replaceVarWithValue(rvar, value);
				i++;
			}
			
			//System.err.println("rvarValuation: "+rvarValuation);
			//System.err.println("etaL:"+etaL);
			//System.err.println("etaLprime:"+etaLprime);
			
			Set<Poly> D1conditions = new HashSet<Poly>();
			
			//add a <= etaLprime - etaL
			//equivalently etaLprime - etaL - a >= 0
			Poly aPoly = new Poly("a");
			
			Poly c1 = PolyUtil.subtract(etaLprime, etaL);
			c1 = PolyUtil.subtract(c1, aPoly);
			//c1>=0
			D1conditions.add(c1);
			
			//add etaLprime-etaL <= b
			//equivalently b - etaLprime + etaL >= 0 
			Poly bPoly = new Poly("b");
			
			Poly c2=PolyUtil.subtract(bPoly, etaLprime);
			c2.addPoly(etaL);
			//c2>=0
			D1conditions.add(c2);
			
			//farkasUtil.addFarkasConditions(invariantConditions, D1conditions);
			allPostConditions.addAll(D1conditions);
		}
		
		//add the type 2 condition
		//E[etaLprime] <= etaL - epsilon
		//compute E[etaLprime]
		Poly ExpEtaLprime = new Poly();
		for(Vector<Integer> rvarValuation:allPossibleValues)
		{
			double prob = Parser.getProbabilityOfValuation(rvarValuation, rvars);
			Poly etaLprime = eta.get(nextStmt).deepCopy();
			etaLprime.replaceVarWithPoly(lhsVar, rhsPoly);
			int i=0;
			for(String rvar:rvars)
			{
				int value = rvarValuation.elementAt(i);
				etaLprime.replaceVarWithValue(rvar, value);
				i++;
			}
			etaLprime.multiplyByConstant(prob);
			ExpEtaLprime.addPoly(etaLprime);
		}
		//end of computing ExpEtaLprime
		Set<Poly> D1BConditions = new HashSet<Poly>();
		//ExpEtaLprime <= etaL - epsilon
		//Equivalently etaL - epsilon - ExpEtaLprime >= 0
		Poly etaL = eta.get(stmt).deepCopy();
		Poly c3 = new Poly();
		c3.addPoly(etaL);
		c3 = PolyUtil.subtract(c3, new Poly("epsilon"));
		c3 = PolyUtil.subtract(c3, ExpEtaLprime);
		D1BConditions.add(c3);
		allPostConditions.addAll(D1BConditions);
		
		//add using Farkas
		farkasUtil.addFarkasConditions(invariantConditions, allPostConditions);
	}
	
	private static void addD2(Node stmt) throws Exception
	{
		
		System.err.println("Processing D2 conditions @"+stmt.statementId+":");
		System.err.println("========================================");
		
		Set<Poly> phi = stmt.children.firstElement().getPolySetOfBexpr();
		Set<Set<Poly>> notPhi = PolyUtil.negationOf(phi);
		Set<Poly> invariant = stmt.invariant.getPolySetOfBexpr();
		Node trueNext = CFGUtil.getTrueNextOf(stmt);
		Node falseNext = CFGUtil.getFalseNextOf(stmt);
		
		//conditions for the true edge
		Set<Poly> truePreconditions = new HashSet<Poly>();
		Set<Poly> truePostConditions = new HashSet<Poly>();
		truePreconditions.addAll(invariant);
		truePreconditions.addAll(phi);
		
		Poly etaL = eta.get(stmt).deepCopy();
		Poly etaLtrue = eta.get(trueNext).deepCopy();
		//a <= etaLtrue - etaL
		//equivalently etaLtrue - etaL - a >= 0
		Poly c1 = PolyUtil.subtract(etaLtrue, etaL);
		c1 = PolyUtil.subtract(c1, new Poly("a"));
		truePostConditions.add(c1);
		
		//etaLtrue - etaL <= b
		//equivalently b - etaLtrue + etaL >= 0
		Poly c2 = new Poly("b");
		c2 = PolyUtil.subtract(c2, etaLtrue);
		c2.addPoly(etaL);
		truePostConditions.add(c2);
		
		//etaLtrue - etaL <= -epsilon
		//equivalently -epsilon - etaLtrue + etaL >= 0
		Poly c3 = new Poly("epsilon");
		c3.multiplyByConstant(-1);
		c3 = PolyUtil.subtract(c3, etaLtrue);
		c3.addPoly(etaL);
		truePostConditions.add(c3);
		
		//add by Farkas
		farkasUtil.addFarkasConditions(truePreconditions, truePostConditions);
		
		//conditions for the false edge
		Poly etaLFalse = eta.get(falseNext).deepCopy();
		for(Set<Poly> cond:notPhi)
		{
			Set<Poly> falsePreconditions = new HashSet<>();
			Set<Poly> falsePostconditions = new HashSet<>();
			
			falsePreconditions.addAll(invariant);
			falsePreconditions.addAll(cond);
			
			//a <= etaLfalse - etaL
			//equivalently etaLfalse - etaL -a >=0 
			Poly d1 = PolyUtil.subtract(etaLFalse, etaL);
			d1 = PolyUtil.subtract(c1, new Poly("a"));
			falsePostconditions.add(d1);
			
			
			//etaLfalse - etaL <= b
			//equivalently b - etaLfalse + etaL >=0
			Poly d2 = new Poly("b");
			d2 = PolyUtil.subtract(d2, etaLFalse);
			d2.addPoly(etaL);
			falsePostconditions.add(d2);
			
			
			//etaLfalse - etaL <= -epsilon
			//equivalently -epsilon - etaLfalse + etaL >= 0
			Poly d3 = new Poly("epsilon");
			d3.multiplyByConstant(-1);
			d3 = PolyUtil.subtract(d3, etaLFalse);
			d3.addPoly(etaL);
			falsePostconditions.add(d3);
			
			
			//add by Farkas
			farkasUtil.addFarkasConditions(falsePreconditions, falsePostconditions);
			
		}
		
	}
	
	private static void addD3(Node stmt) throws Exception
	{
		
		System.err.println("Processing D3 conditions @"+stmt.statementId+":");
		System.err.println("========================================");
		
		Set<Node> nextStatements = CFGUtil.getSuccessorsOf(stmt);
		Set<Poly> invariant = stmt.invariant.getPolySetOfBexpr();
		Poly etaL = eta.get(stmt).deepCopy();
		
		for(Node next:nextStatements)
		{
			Poly etaLprime = eta.get(next).deepCopy();
			
			Set<Poly> postconditions = new HashSet<>();
			
			//a <= etaLprime - etaL
			//Equivalently etaLprime - etaL - a >= 0
			Poly c1 = PolyUtil.subtract(etaLprime, etaL);
			c1 = PolyUtil.subtract(c1, new Poly("a"));
			postconditions.add(c1);
			

			//etaLprime - etaL <= b
			//equivalently b - etaLprime + etaL >= 0
			Poly c2 = new Poly("b");
			c2 = PolyUtil.subtract(c2, etaLprime);
			c2.addPoly(etaL);
			postconditions.add(c2);
			
			//etaLprime - etaL <= -epsilon
			//equivalently -epsilon - etaLprime + etaL >= 0
			Poly c3 = new Poly("epsilon");
			c3.multiplyByConstant(-1);
			c3 = PolyUtil.subtract(c3, etaLprime);
			c3.addPoly(etaL);
			postconditions.add(c3);
			
			//add by Farkas
			farkasUtil.addFarkasConditions(invariant, postconditions);
		}
		
	}
	
	private static void addD4(Node stmt) throws Exception
	{
		System.err.println("Processing D4 conditions @"+stmt.statementId+":");
		System.err.println("========================================");
		
		Set<Node> nextStatements = CFGUtil.getSuccessorsOf(stmt);
		Set<Poly> invariant = stmt.invariant.getPolySetOfBexpr();
		Poly etaL = eta.get(stmt).deepCopy();
		Set<Poly> postconditions = new HashSet<>();
		
		//first type of condition
		for(Node next:nextStatements)
		{
			Poly etaLprime = eta.get(next).deepCopy();
			
			//a<=etaLprime - etaL
			//equivalently etaLprime - etaL - a >=0
			Poly c1 = PolyUtil.subtract(etaLprime, etaL);
			c1 = PolyUtil.subtract(c1, new Poly("a"));
			postconditions.add(c1);
			
			//etaLprime - etaL <= b
			//equivalently b - etaLprime + etaL >= 0
			Poly c2 = new Poly("b");
			c2 = PolyUtil.subtract(c2, etaLprime);
			c2.addPoly(etaL);
			postconditions.add(c2);
			
		}
		
		//second type of condition (weighted average)
		double p = stmt.children.firstElement().getValue();
		Node trueNext = CFGUtil.getTrueNextOf(stmt);
		Node falseNext = CFGUtil.getFalseNextOf(stmt);
		Poly etaTrue = eta.get(trueNext).deepCopy();
		Poly etaFalse = eta.get(falseNext).deepCopy();
		
		//p * etaTrue + (1-p) * etaFalse <= etaL - epsilon
		//Equivalently
		//etaL - epsilon - p * etaTrue - (1-p) * etaFalse >= 0
		Poly d = PolyUtil.subtract(etaL, new Poly("epsilon"));
		etaTrue.multiplyByConstant(p);
		etaFalse.multiplyByConstant(1-p);
		d = PolyUtil.subtract(d, etaTrue);
		d = PolyUtil.subtract(d, etaFalse);
		postconditions.add(d);
		
		
		//add by Farkas
		farkasUtil.addFarkasConditions(invariant, postconditions);
		
	}
	
	private static void addD5(Node stmt) throws Exception
	{
		System.err.println("Processing D5 conditions @"+stmt.statementId+":");
		System.err.println("========================================");
		
		Set<Poly> G = stmt.children.firstElement().getPolySetOfBexpr();
		Set<Poly> invariant = stmt.invariant.getPolySetOfBexpr();
		Poly etaL = eta.get(stmt).deepCopy();
		
		Set<Poly> preconditions = new HashSet<>();
		Set<Poly> postconditions = new HashSet<>();
		
		preconditions.addAll(G);
		preconditions.addAll(invariant);
		
		//etaL >= c
		//equivalently etaL - c >= 0
		Poly p = PolyUtil.subtract(etaL, new Poly("c"));
		postconditions.add(p);
		
		//add by Farkas
		farkasUtil.addFarkasConditions(preconditions, postconditions);
	}

	public static void printDSM(Result result) {
		System.err.println("The final result is:");
		System.err.println("========================================");
		if(result==null)
		{
			System.err.println("DSM Synthesis Failed.");
			return;
		}
		System.err.println("DSM with epsilon="+result.get("epsilon")+"  c="+result.get("c")+"  [a, b] = ["+result.get("a")+", "+result.get("b")+"]");
		
		Set<String> evars = new TreeSet<String>();
		for(int i=1;i<eCnt;++i)
			evars.add("e"+i);
		
		for(Node stmt:Node.allStatements)
		{
			Poly stmtEta = eta.get(stmt).deepCopy();
			for(String evar:evars)
				stmtEta.replaceVarWithValue(evar, result.get(evar).doubleValue());
			System.err.println("eta("+stmt.statementId+") = "+stmtEta);
		}
	}
	
}
