package DSM;

import java.util.Vector;

import net.sf.javailp.Linear;

//models a linear expression
public class Poly {
	public Vector<PolyTerm> terms;
	public Poly()
	{
		terms = new Vector<>();
	}
	
	Poly(PolyTerm t)
	{
		terms = new Vector<>();
		addTerm(t);
	}
	
	Poly(String var)
	{
		terms = new Vector<>();
		addTerm(new PolyTerm(var));
	}
	
	public String toString()
	{
		removeZeros();
		String ans = "";
		boolean firstTerm = true;
		for(PolyTerm t:terms)
		{
			if(!firstTerm && t.constant>0)
				ans+=" + ";
			ans+=t.toString();
			firstTerm = false;
		}
		if(terms.size()==0)
			ans+="0";
		return ans;
	}
	
	Poly deepCopy()
	{
		Poly ans = new Poly();
		ans.addPoly(this);
		ans.removeZeros();
		return ans;
	}
	
	void addPoly(Poly poly)
	{
		for(PolyTerm t:poly.terms)
			addTerm(t);
	}
	
	public void addTerm(PolyTerm term)
	{
		term = term.deepCopy();
		boolean added = false;
		for(PolyTerm t:terms)
			if(t.hasSamePowersAs(term))
			{
				t.constant += term.constant;
				added=true;
				break;
			}
		if(!added)
			terms.add(term);
		removeZeros();
	}
	
	void removeZeros()
	{
		double EPS = 1e-8;
		Vector<PolyTerm> remove = new Vector<>();
		for(PolyTerm t:terms)
			if(t.constant<EPS && t.constant>-EPS)
				remove.add(t);
		for(PolyTerm t:remove)
			terms.remove(t);
	}

	public void multiplyByConstant(double cns) {
		for(PolyTerm t:terms)
			t.multiplyByConstant(cns);
		removeZeros();
	}

	public void replaceVarWithValue(String rvar, double value) {
		Poly ans = new Poly();
		for(PolyTerm t:terms)
		{
			PolyTerm newT = t.deepCopy();
			newT.replaceVarWithValue(rvar, value);
			ans.addTerm(newT);
		}
		terms = ans.terms;
	}

	public void replaceVarWithPoly(String lhs, Poly rhs) {
		Poly ans = new Poly();
		
		for(PolyTerm t:terms)
		{
			Poly novoT = t.replaceVarWithPoly(lhs, rhs);
			ans.addPoly(novoT);
		}
		
		terms = ans.terms;
	}

	public void multiplyByVarPower(String var, int power) {
		for(PolyTerm t:terms)
			t.multiplyByVarPower(var, power);
	}

	public void multiplyByPoly(Poly p) {
		Poly ans = new Poly();
		
		for(PolyTerm t:p.terms)
		{
			Poly add = this.deepCopy();
			add.multiplyByTerm(t);
			ans.addPoly(add);
		}
		
		terms = ans.terms;
	}

	private void multiplyByTerm(PolyTerm term) {
		for(PolyTerm t:terms)
			t.multiplyByTerm(term);
	}

	public Linear toLinear() throws Exception { //assumes the poly is linear
		Linear ans = new Linear();
		for(PolyTerm t:terms)
		{
			if(t.powers.size()==0)
				ans.add(t.constant,"oneVar");
			else if(t.powers.size()==1 && t.getPowerOf(t.powers.firstKey())==1)
				ans.add(t.constant, t.powers.firstKey());
			else
				throw new Exception("toLinear is called for non-linear poly "+this);
		}
		return ans;
	}
}
