package DSM;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

public class PolyTerm {
	public TreeMap<String, Integer> powers;
	public double constant;
	
	public PolyTerm()
	{
		constant = 0;
		powers = new TreeMap<>();
	}
	
	PolyTerm(String var)
	{
		constant = 1;
		powers = new TreeMap<>();
		powers.put(var, 1);
	}
	
	public PolyTerm deepCopy()
	{
		PolyTerm ans = new PolyTerm();
		ans.constant = this.constant;
		for(String var:powers.keySet())
			ans.powers.put(var, getPowerOf(var));
		ans.removeZeros();
		return ans;
	}
	
	public String toString()
	{
		double EPS = 1e-8;
		removeZeros();
		if(constant<EPS && constant>-EPS)
			return "0";
		String ans = "";
		if(constant<1+EPS && constant>1-EPS)
		{
			//do nothing -- the constant is 1
			if(powers.isEmpty())
				ans+="1";
		}
		else if(constant<-1+EPS && constant>-1-EPS)
		{
			//the constant is -1
			ans+="-";
			if(powers.isEmpty())
				ans+="1";
		}
		else
		{
			ans+=constant+" ";
		}
		for(String var:powers.keySet())
		{
			ans+=var;
			if(getPowerOf(var)!=1)
				ans+="^"+getPowerOf(var);
			ans+=" ";
		}
		return ans;
	}

	public boolean hasSamePowersAs(PolyTerm term) {
		removeZeros();
		term.removeZeros();
		for(String var:powers.keySet())
			if(term.getPowerOf(var)!=getPowerOf(var))
				return false;
		for(String var:term.powers.keySet())
			if(term.getPowerOf(var)!=getPowerOf(var))
				return false;
		return true;
	}
	
	void removeZeros()
	{
		Set<String> remove = new HashSet<>();
		for(String var:powers.keySet())
			if(getPowerOf(var)==0)
				remove.add(var);
		
		for(String var:remove)
			powers.remove(var);
	}
	
	public int getPowerOf(String var)
	{
		if(powers.containsKey(var))
			return powers.get(var);
		return 0;
	}

	public void multiplyByConstant(double cns) {
		constant *= cns;		
	}

	public void replaceVarWithValue(String rvar, double value) {
		int pw = getPowerOf(rvar);
		for(int i=0;i<pw;++i)
			constant *= value;
		powers.remove(rvar);
	}

	public Poly replaceVarWithPoly(String lhs, Poly rhs) {
		Poly ans = new Poly();
		PolyTerm one = new PolyTerm();
		one.constant = 1;
		ans.addTerm(one);
		//ans = 1
		
		//apply the var^powers
		for(String var:powers.keySet())
		{
			int power = getPowerOf(var);
			if(!var.equals(lhs))
				ans.multiplyByVarPower(var, power);
			else
			{
				for(int i=0;i<power;++i)
					ans.multiplyByPoly(rhs);
			}
		}
		
		//apply the constant
		ans.multiplyByConstant(constant);
		
		return ans;
	}

	public void multiplyByVarPower(String var, int power) {
		powers.put(var, getPowerOf(var)+power);
	}

	public void multiplyByTerm(PolyTerm term) {
		//constants
		multiplyByConstant(term.constant);
		
		//others
		for(String var:term.powers.keySet())
			multiplyByVarPower(var, term.getPowerOf(var));
	}
	
}
