package main;

import java.time.LocalTime;
import static java.time.temporal.ChronoUnit.MILLIS;

import CFG.CFGUtil;
import DSM.DSMUtil;
import farkas.farkasUtil;
import net.sf.javailp.Result;
import parser.Node;
import parser.Parser;

public class Main {
	
	public static void main(String[] args) throws Exception
	{
		LocalTime beginTime = LocalTime.now();
		System.err.println("========================================");
		System.err.println("Analysis Started at "+beginTime);
		
		//Parse
		if(args.length<1)
			throw new Exception("Please specify the input file.");
		String filePath = args[0];
		Parser.readFile(filePath);
		Node root = Parser.parseProg(null, 0, Parser.getNumberOfTokens()-1);
		
		System.err.println("Parsing is done. Parse tree:");
		System.err.println("========================================");
		root.print();
		System.err.println("========================================");
		
		//Make CFG
		CFGUtil.addEdges(root);
		
		System.err.println("Control Flow Graph:");
		System.err.println("========================================");
		for(CFG.Edge e:CFG.Edge.allEdges)
			System.err.println(e);
		System.err.println("========================================");
		
		//Initialize the LPSolver
		farkasUtil.initializeJavaILP();
		
		//DSM
		DSMUtil.createEtaTemplates();
		DSMUtil.addDSMConditions();
		
		//LP Solver
		Result result = farkasUtil.solveLP();
		
		//Print the final DSM
		DSMUtil.printDSM(result);
		
		LocalTime endTime = LocalTime.now();
		System.err.println("========================================");
		System.err.println("Analysis Finished at "+endTime);
		System.err.println("Total runtime (in milliseconds):"+beginTime.until(endTime, MILLIS));
	}

}
