package parser;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

public class Parser {

	//global variables
	static Vector<String> tokens = new Vector<>();
	static Map<String, Map<Integer, Double>> samplingVariablesSupport = new HashMap<>();
	
	public static Node parseStmt(Node parent, int beginIndex, int endIndex) throws Exception
	{
		Node invariant = null;
		int firstSharpIndex = beginIndex;
		int secondSharpIndex = -1;
		//check for invariant
		if(getToken(beginIndex).equals("#")) //there is an invariant
		{
			for(int i=beginIndex+1;i<=endIndex;++i)
				if(getToken(i).equals("#"))
				{
					secondSharpIndex = i;
					break;
				}
			if(secondSharpIndex==-1)
				throw new Exception("No end point found for the invariant beginning @"+firstSharpIndex);
			
		}
		else
			throw new Exception("Missing invariant for the statement @["+beginIndex+", "+endIndex+"]");
		
		beginIndex = secondSharpIndex+1;
		
		//end of checking for invariant
			
		if(getToken(beginIndex).equals("skip")) //skip statement
		{
			if(endIndex!=beginIndex)
				throw new Exception("beginIndex!=endIndex @skip statement: ["+beginIndex+","+endIndex+"]");
			Node skipNode = new Node(parent, "stmt", "skip", beginIndex, endIndex);
			//fix the invariant
			invariant = parseBexpr(skipNode, firstSharpIndex+1, secondSharpIndex-1);
			skipNode.children.removeElement(invariant);
			skipNode.invariant = invariant;
			return skipNode;
		}
		else if(getToken(beginIndex).equals("while")) //while statement
		{
			if(!getToken(endIndex).equals("od"))
				throw new Exception("Expected od @"+endIndex+" for the while @"+beginIndex);
			int whileIndex = beginIndex;
			int odIndex = endIndex;
			//find the first do
			int doIndex = -1;
			for(int i=beginIndex;i<=endIndex;++i)
				if(getToken(i).equals("do"))
				{
					doIndex = i;
					break;
				}
			if(doIndex==-1)
				throw new Exception("Did not find a do for the while @"+whileIndex);
			
			Node whileNode = new Node(parent, "stmt", "while", beginIndex, endIndex);
			
			parseBexpr(whileNode, whileIndex+1, doIndex-1);
			parseProg(whileNode, doIndex+1, odIndex-1);
			
			//fix the invariant
			invariant = parseBexpr(whileNode, firstSharpIndex+1, secondSharpIndex-1);
			whileNode.children.removeElement(invariant);
			whileNode.invariant = invariant;
			
			return whileNode;
			
		}
		else if(getToken(beginIndex).equals("if") && getToken(beginIndex+1).equals("*")) //if* statement
		{
			
			int ifIndex = beginIndex;
			int starIndex = beginIndex+1;
			if(!getToken(beginIndex+2).equals("then"))
				throw new Exception("Expected then @"+beginIndex+2+" for the if @"+ifIndex);
			int thenIndex = beginIndex+2;
			if(!getToken(endIndex).equals("fi"))
				throw new Exception("Expected fi @"+endIndex+" for the if @"+ifIndex);
			int fiIndex = endIndex;
			//find the then
			int elseIndex = getElseOfIf(beginIndex, endIndex);
			
			Node ifStarNode = new Node(parent, "stmt", "if*", beginIndex, endIndex);
			
			//then part
			parseProg(ifStarNode, thenIndex+1, elseIndex-1);
			
			//else part
			parseProg(ifStarNode, elseIndex+1, fiIndex-1);
			
			//fix the invariant
			invariant = parseBexpr(ifStarNode, firstSharpIndex+1, secondSharpIndex-1);
			ifStarNode.children.removeElement(invariant);
			ifStarNode.invariant = invariant;
			
			return ifStarNode;
			
		}
		else if(getToken(beginIndex).equals("if") && getToken(beginIndex+1).equals("prob")) //ifprob statement
		{
			int ifIndex = beginIndex;
			int probIndex = beginIndex+1;
			if(!getToken(probIndex+1).equals("(") || !getToken(probIndex+3).equals(")"))
				throw new Exception("Invalid parenthesization for the ifprob @"+ifIndex);
			int valueIndex = probIndex + 2;
			int thenIndex = valueIndex + 2;
			if(!getToken(thenIndex).equals("then"))
				throw new Exception("Expected a then @"+thenIndex+" for the if @"+ifIndex);
			int fiIndex = endIndex;
			if(!getToken(fiIndex).equals("fi"))
				throw new Exception("Expected a fi @"+fiIndex+" for the if @"+ifIndex);
			int elseIndex = getElseOfIf(beginIndex, endIndex);
			
			Node ifProbNode = new Node(parent, "stmt", "ifprob", beginIndex, endIndex);
			
			//constant
			parseConstant(ifProbNode, valueIndex, valueIndex);
			
			//then part
			parseProg(ifProbNode, thenIndex+1, elseIndex-1);
			
			//else part
			parseProg(ifProbNode, elseIndex+1, fiIndex-1);
			
			//fix the invariant
			invariant = parseBexpr(ifProbNode, firstSharpIndex+1, secondSharpIndex-1);
			ifProbNode.children.removeElement(invariant);
			ifProbNode.invariant = invariant;
			
			return ifProbNode;
		}
		else if(getToken(beginIndex).equals("if"))//if statement
		{
			int ifIndex = beginIndex;
			int fiIndex = endIndex;
			if(!getToken(fiIndex).equals("fi"))
				throw new Exception("Expected fi @"+fiIndex+" for the if @"+ifIndex);
			int thenIndex = -1;
			for(int i=beginIndex;i<=endIndex;++i)
				if(getToken(i).equals("then"))
				{
					thenIndex = i;
					break;
				}
			if(thenIndex==-1)
				throw new Exception("Could not find the then for the if @"+ifIndex);
			int elseIndex = getElseOfIf(beginIndex, endIndex);
			
			Node ifNode = new Node(parent, "stmt", "if", beginIndex, endIndex);
			
			//bexpr
			parseBexpr(ifNode, ifIndex+1, thenIndex-1);
			
			//then part
			parseProg(ifNode, thenIndex+1, elseIndex-1);
			
			//else part
			parseProg(ifNode, elseIndex+1, fiIndex-1);
			
			//fix the invariant
			invariant = parseBexpr(ifNode, firstSharpIndex+1, secondSharpIndex-1);
			ifNode.children.removeElement(invariant);
			ifNode.invariant = invariant;
			
			return ifNode;
		}
		else //assignment statement 
		{
			//find the :=
			int eqIndex = beginIndex+1;
			if(!getToken(eqIndex).equals(":="))
				throw new Exception("Expected := @"+eqIndex);
			
			Node assignmentNode = new Node(parent, "stmt", "assignment", beginIndex, endIndex);
			
			//lhs
			Node lhs = parseVar(assignmentNode, beginIndex, beginIndex);
			if(!lhs.rule.equals("pvar"))
				throw new Exception("Expected a pvar as the lhs of the assignment @["+beginIndex+" ,"+endIndex+"]");
			
			//rhs
			parseExpr(assignmentNode, eqIndex+1, endIndex);
			
			//fix the invariant
			invariant = parseBexpr(assignmentNode, firstSharpIndex+1, secondSharpIndex-1);
			assignmentNode.children.removeElement(invariant);
			assignmentNode.invariant = invariant;
			
			return assignmentNode;
		}
	}
	
	private static Node parseExpr(Node parent, int beginIndex, int endIndex) throws Exception
	{
		//find all +/-'s
		Vector<Integer> opIndices = new Vector<>();
		Vector<Character> ops = new Vector<>();
		//the first +
		if(!getToken(beginIndex).equals("-"))
		{
			opIndices.add(beginIndex-1);
			ops.add('+');
		}
		//all others
		for(int i=beginIndex;i<=endIndex;++i)
			if(getToken(i).equals("+") || getToken(i).equals("-"))
			{
				opIndices.add(i);
				ops.add(getToken(i).charAt(0));
			}
		//the last (dummy) +
		opIndices.add(endIndex+1);
		ops.add('+');
		
		//create a new node
		Node expr = new Node(parent, "expr", "+", beginIndex, endIndex);
		
		//add all the terms
		int lastOpIndex = opIndices.firstElement();
		for(int i=1;i<opIndices.size();++i)
		{
			int nextOpIndex = opIndices.elementAt(i);
			
			Node term = parseTerm(expr, lastOpIndex+1, nextOpIndex-1);
			if(ops.elementAt(i-1)=='-')
			{
				//make the term negative
				Node cons = term.children.elementAt(0);
				cons.setValue(-cons.getValue());
			}
			
			lastOpIndex = nextOpIndex;
		}
		
		//return
		return expr;
	}
	
	private static Node parseTerm(Node parent, int beginIndex, int endIndex) throws Exception {
		if(beginIndex==endIndex && (getToken(beginIndex).startsWith("p") || getToken(beginIndex).startsWith("r"))) //a single variable
		{
			Node term = new Node(parent, "term", "var", beginIndex, endIndex);
			Node cons = new Node(term, "constant", "constant", beginIndex, endIndex);
			cons.setValue(1);
			parseVar(term, beginIndex, endIndex);
			return term;
		}
		else if(beginIndex==endIndex) //a single constant
		{
			Node term = new Node(parent, "term", "constant", beginIndex, endIndex);
			parseConstant(term, beginIndex, endIndex);
			return term;
		}
		else //constant * variable
		{
			if(beginIndex!=endIndex-2 || !getToken(beginIndex+1).equals("*"))
				throw new Exception("Malformed term @["+beginIndex+", "+endIndex+"].");
			Node term = new Node(parent, "term", "*", beginIndex, endIndex);
			parseConstant(term, beginIndex, beginIndex);
			parseVar(term, beginIndex+2, beginIndex+2);
			return term;
		}
	}

	private static Node parsePterm(Node parent, int beginIndex, int endIndex) throws Exception {
		if(beginIndex==endIndex && (getToken(beginIndex).startsWith("p") || getToken(beginIndex).startsWith("r"))) //a single variable
		{
			Node pterm = new Node(parent, "pterm", "pvar", beginIndex, endIndex);
			Node cons = new Node(pterm, "constant", "constant", beginIndex, endIndex);
			cons.setValue(1);
			Node pvar = parseVar(pterm, beginIndex, endIndex);
			if(!pvar.rule.equals("pvar"))
				throw new Exception("Only pvars are allowed @"+beginIndex);
			return pterm;
		}
		else if(beginIndex==endIndex) //a single constant
		{
			Node pterm = new Node(parent, "pterm", "constant", beginIndex, endIndex);
			parseConstant(pterm, beginIndex, endIndex);
			return pterm;
		}
		else //constant * variable
		{
			if(beginIndex!=endIndex-2 || !getToken(beginIndex+1).equals("*"))
				throw new Exception("Malformed term @["+beginIndex+", "+endIndex+"].");
			Node pterm = new Node(parent, "pterm", "*", beginIndex, endIndex);
			parseConstant(pterm, beginIndex, beginIndex);
			Node pvar = parseVar(pterm, beginIndex+2, beginIndex+2);
			if(!pvar.rule.equals("pvar"))
				throw new Exception("Only pvars are allowed @"+beginIndex+2);
			return pterm;
		}
	}
	
	private static Node parsePexpr(Node parent, int beginIndex, int endIndex) throws Exception
	{
		//find all +/-'s
		Vector<Integer> opIndices = new Vector<>();
		Vector<Character> ops = new Vector<>();
		//the first +
		if(!getToken(beginIndex).equals("-"))
		{
			opIndices.add(beginIndex-1);
			ops.add('+');
		}
		//all others
		for(int i=beginIndex;i<=endIndex;++i)
			if(getToken(i).equals("+") || getToken(i).equals("-"))
			{
				opIndices.add(i);
				ops.add(getToken(i).charAt(0));
			}
		//the last (dummy) +
		opIndices.add(endIndex+1);
		ops.add('+');
		
		//create a new node
		Node pexpr = new Node(parent, "pexpr", "+", beginIndex, endIndex);
		
		//add all the terms
		int lastOpIndex = opIndices.firstElement();
		for(int i=1;i<opIndices.size();++i)
		{
			int nextOpIndex = opIndices.elementAt(i);
			
			Node pterm = parsePterm(pexpr, lastOpIndex+1, nextOpIndex-1);
			if(ops.elementAt(i-1)=='-')
			{
				//make the term negative
				Node cons = pterm.children.elementAt(0);
				cons.setValue(-cons.getValue());
			}
			
			lastOpIndex = nextOpIndex;
		}
		
		//return
		return pexpr;
	}
	
	private static Node parseBexpr(Node parent, int beginIndex, int endIndex) throws Exception {
		//find all the or's
		Vector<Integer> ors = new Vector<>();
		for(int i=beginIndex;i<=endIndex;++i)
			if(getToken(i).equals("or"))
				ors.add(i);
		
		//create a node
		Node bexpr = new Node(parent, "bexpr", "or", beginIndex, endIndex);
		
		//add all the andbexprs
		int lastOrIndex = beginIndex - 1;
		ors.add(endIndex+1);
		for(int orIndex:ors)
		{
			parseAndbexpr(bexpr, lastOrIndex+1, orIndex-1);
			lastOrIndex = orIndex;
		}
		
		//return the node
		return bexpr;
	}
	
	private static Node parseAndbexpr(Node parent, int beginIndex, int endIndex) throws Exception
	{
		//find all the and's
		Vector<Integer> ands = new Vector<>();
		for(int i=beginIndex; i<=endIndex;++i)
			if(getToken(i).equals("and"))
				ands.add(i);
		
		//create a node
		Node andbexpr = new Node(parent, "andbexpr", "and", beginIndex, endIndex);
		
		//add all the literals
		int lastAndIndex = beginIndex - 1;
		ands.add(endIndex+1);
		for(int andIndex:ands)
		{
			parseLiteral(andbexpr, lastAndIndex+1, andIndex-1);
			lastAndIndex = andIndex;
		}
		
		//return the node
		return andbexpr;
	}
	
	private static Node parseLiteral(Node parent, int beginIndex, int endIndex) throws Exception
	{
		//find the operator
		int opIndex = -1;
		for(int i=beginIndex; i<=endIndex;++i)
			if(getToken(i).equals("<=") || getToken(i).equals(">="))
			{
				opIndex = i;
				break;
			}
		if(opIndex==-1)
			throw new Exception("Could not find an operator <= or >= in the literal @["+beginIndex+", "+endIndex+"]");
		
		//create a node
		Node literal = new Node(parent, "literal", getToken(opIndex), beginIndex, endIndex);
		
		
		//add lhs
		parsePexpr(literal, beginIndex, opIndex-1);
		
		//add rhs
		parsePexpr(literal, opIndex+1, endIndex);
		
		//System.err.println("Returned literal is:"+literal);
		
		//return it
		return literal;
	}

	static int getElseOfIf(int beginIndex, int endIndex) throws Exception
	{
		int elseIndex = -1;
		int openIfs = 0;
		for(int i=beginIndex;i<=endIndex;++i)
		{
			if(getToken(i).equals("if"))
				openIfs++;
			else if(getToken(i).equals("fi"))
			{
				openIfs--;
				if(openIfs<0)
					throw new Exception("Unexpected fi @"+i);
			}
			else if(getToken(i).equals("else") && openIfs==1)
				elseIndex = i;
		}
		
		if(elseIndex==-1)
			throw new Exception("Could not find an else for the if @"+beginIndex);
		
		return elseIndex;
	}
	
	private static Node parseVar(Node parent, int beginIndex, int endIndex) throws Exception
	{
		if(beginIndex!=endIndex)
			throw new Exception("beginIndex!=endIndex for the var @["+beginIndex+", "+endIndex+"]");
		String varName = getToken(beginIndex);
		if(!varName.startsWith("p") && !varName.startsWith("r"))
			throw new Exception(varName+" @"+beginIndex+" is not a valid variable name.");
		
		Node var = null;
		if(varName.startsWith("p"))
			var = new Node(parent, "var", "pvar", beginIndex, endIndex);
		else
			var = new Node(parent, "var", "rvar", beginIndex, endIndex);
		var.setName(varName);
		return var;
	}
	
	private static Node parseConstant(Node parent, int beginIndex, int endIndex) throws Exception
	{
		if(beginIndex!=endIndex)
			throw new Exception("beginIndex!=endIndex for the constant @["+beginIndex+", "+endIndex+"]");
		Node constant = new Node(parent, "constant", "constant", beginIndex, endIndex);
		constant.setValue(Double.parseDouble(getToken(beginIndex)));
		return constant;
	}
	
	public static Node parseProg(Node parent, int beginIndex, int endIndex) throws Exception
	{
		//find all semicolons
		Vector<Integer> semicolons = new Vector<>();
		
		int openIfs = 0;
		int openWhiles = 0;
		for(int i=beginIndex;i<=endIndex;++i)
		{
			if(getToken(i).equals("if"))
				openIfs++;
			else if(getToken(i).equals("while"))
				openWhiles++;
			else if(getToken(i).equals("fi"))
				openIfs--;
			else if(getToken(i).equals("od"))
				openWhiles--;
			
			if(openIfs<0 || openWhiles<0)
				throw new Exception("@token "+i+": unopened if or while is closed.");
			
			if(openIfs!=0 || openWhiles!=0)
				continue;
			else if(getToken(i).equals(";"))
				semicolons.add(i);
		}
		
		//create the new node
		Node prog = new Node(parent, "prog", "prog", beginIndex, endIndex);
		
		//parse each statement
		semicolons.add(endIndex+1);
		int lastSemicolon = beginIndex - 1;
		for(int semicolon:semicolons)
		{
			parseStmt(prog, lastSemicolon+1, semicolon-1);
			lastSemicolon = semicolon;
		}
		
		//return the node
		return prog;
	}
	
	public static String getToken(int i)
	{
		return tokens.elementAt(i);
	}
	
	public static int getNumberOfTokens()
	{
		return tokens.size();
	}
	
	static void readTokensFrom(String input) throws Exception //fixes the spaces for parsing and then tokenize the input
	{
		String extraSpace = "";
		for(int i=0;i<input.length();++i)
		{
			char c = input.charAt(i);
			if(c=='.' || Character.isAlphabetic(c) || Character.isDigit(c))
				extraSpace += c;
			else
			{
				extraSpace += " ";
				extraSpace += c;
				extraSpace += " ";
			}
		}
		
		Scanner cin = new Scanner(extraSpace);
		
		while(cin.hasNext())
		{
			String x = cin.next();
			if(!x.equals("="))
				tokens.add(x);
			else
			{
				if(tokens.size()==0)
					throw new Exception("A program cannot start with =");
				String y = tokens.lastElement();
				y += x;
				tokens.removeElementAt(tokens.size()-1);
				tokens.addElement(y);
			}
		}
		
		cin.close();
		
	}
	
	public static void readFile(String filePath) throws Exception
	{
		File file = new File(filePath);
		Scanner cin = new Scanner(file);
		
		//read the input program
		String inputProgram = "";
		while(cin.hasNextLine())
		{
			String line = cin.nextLine();
			if(line.equals("end"))
				break;
			inputProgram += " "+line;
		}
		readTokensFrom(inputProgram);
		
		//print the tokens
		System.err.println("========================================");
		System.err.println("The tokens are:");
		System.err.println("========================================");
		for(int i=0;i<tokens.size();++i)
			System.err.println(i+":"+tokens.elementAt(i));
		System.err.println("========================================");
		
		//read the support of sampling variables
		while(cin.hasNextLine())
		{
			String line = cin.nextLine();
			Scanner rvarScan = new Scanner(line);
			String rvar = rvarScan.next();
			samplingVariablesSupport.put(rvar, new HashMap<>());
			while(rvarScan.hasNext())
			{
				int value = rvarScan.nextInt();
				double probability = rvarScan.nextDouble();
				samplingVariablesSupport.get(rvar).put(value, probability);
			}
			rvarScan.close();
		}
		
		//print the supports
		System.err.println("The sampling variables and their distributions are:");
		System.err.println("========================================");
		System.err.println(samplingVariablesSupport);
		System.err.println("========================================");
		
		cin.close();
	}

	public static Set<Vector<Integer>> getAllPossibleCombinationsOfValuesOfRvars(TreeSet<String> rvars) { //returns a set of vectors, each containing one possible valuation to the sampling variables in the set rvars
		if(rvars.size()==0)
		{
			HashSet<Vector<Integer>> ans = new HashSet<>();
			ans.add(new Vector<>());
			return ans;
		}
		
		TreeSet<String> smallRvars = new TreeSet<String>();
		smallRvars.remove(rvars.last());
		Set<Vector<Integer>> smallAns = getAllPossibleCombinationsOfValuesOfRvars(smallRvars);
		String newRvar = rvars.last();
		
		HashSet<Vector<Integer>> ans = new HashSet<>();
		for(Vector<Integer> smallVal:smallAns)
			for(int newRvarVal:samplingVariablesSupport.get(newRvar).keySet())
			{
				Vector<Integer> newVal = new Vector<Integer>();
				newVal.addAll(smallVal);
				newVal.add(newRvarVal);
				ans.add(newVal);
			}
		
		return ans;
	
	}

	public static double getProbabilityOfValuation(Vector<Integer> rvarValuation, TreeSet<String> rvars) {
		double ans = 1;
		int i = 0;
		for(String var:rvars)
		{
			int val = rvarValuation.elementAt(i);
			ans *= samplingVariablesSupport.get(var).get(val);
			i++;
		}
		return ans;
	}
}
